// ©2023 Batty Bovine Productions, LLC. All Rights Reserved.

#pragma once

#include "CoreMinimal.h"
#include "EdGraph/EdGraphNode.h"

#include "EdComboActionGraphEdge.generated.h"


/**
 * 
 */
UCLASS(MinimalAPI)
class  UEdComboActionGraphEdge : public UEdGraphNode
{
	GENERATED_BODY()

public:
	UPROPERTY()
		class UEdGraph *Graph;

	UPROPERTY(VisibleAnywhere, Instanced, Category="Combo Action Graph")
		class UComboActionGraphEdge *ComboActionGraphEdge;

public:
	
	void SetEdge(class UComboActionGraphEdge *Edge);

	virtual void AllocateDefaultPins() override;

	virtual FText GetNodeTitle(ENodeTitleType::Type TitleType) const override;

	virtual void PinConnectionListChanged(UEdGraphPin *Pin) override;

	virtual void PrepareForCopying() override;

	virtual UEdGraphPin *GetInputPin() const { return Pins[0]; }
	virtual UEdGraphPin *GetOutputPin() const { return Pins[1]; }

	void CreateConnections(class UEdComboActionGraphNode *Start, class UEdComboActionGraphNode* End);

	class UEdComboActionGraphNode *GetStartNode();
	class UEdComboActionGraphNode *GetEndNode();
};
